/*
 *----------------------------------------------------------------------------
 * Copyright (c) 2001, 2002, C.S. Peron
 * All rights reserved.
 *----------------------------------------------------------------------------
 *
 *  * Redistribution and use in source and binary forms, with or without 
 *    modification, are permitted provided that the following condition
 *    are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above 
 *    copyright notice, this list of conditions and the following disclaimer 
 *    in the documentation and/or other materials provided with
 *    the distribution.
 *
 *  * Neither the name of the ipex developers nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission. 
 *
 *----------------------------------------------------------------------------
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *----------------------------------------------------------------------------
 */

#ifndef lint
static const char rcsid[] =
        "@(#) $Header: /usr/cvs/ipex24000/ipex.c,v 1.13 2004/07/29 03:59:12 modulus Exp $";
#endif

#include <ipex_includes.h>

#define READ_TIMEOUT 1000
#define NO_PRINT 0x7f


static	char	**protocols = NULL;
static	char	**tcp_ports = NULL;
static	char	**udp_ports = NULL;

/*
 * lookup table, the function pointer in
 * delegate will initialize by retrieving the
 * ip_p section of the IP header and using
 * it to cross reference it's arbitrary handler
 * in this table
 */
static const struct proto_printers {
	u_char	pp_proto;	/* ip protocol */
	int	pp_hlen;	/* required header length */

	int (*pp_printer)(u_char *, register struct pcap_pkthdr *, u_char *);
} p_funcs[] = {
	{ IPPROTO_TCP,		24,	print_tcp },
	{ IPPROTO_ICMP,		4,	print_icmp },
	{ IPPROTO_UDP,		8,	print_udp },
	{ IPPROTO_IGMP,		8,	print_igmp },
	{ NO_PRINT,		0,	generic_print },
	{ 0,			0,	NULL }
};

pcap_t *pd = NULL;
pcap_dumper_t *pdmper;
Heap	*conqueue;

int
main(int argc, char *argv [])
{
	char ebuf[PCAP_ERRBUF_SIZE + 32];
	int pid, fg = 1, pidfd, sn, ch, cnt, snaplen = DEFAULT_SNAPLEN;
	glob_t g;
	FILE *cmdfp;
	char *cmdbuf, *fname, *rem;
	u_char *pargs;
	struct in_addr localnet, netmask;
	struct bpf_program bpfcode;
	struct pcap_stat pstat;
	struct dump_info info;

	cmdbuf = NULL;
	if (argc == 2) {
		if (strncmp("--help", argv[1], 6) == 0) {
			usage(NULL);
			exit(EX_USAGE);
		}
	}
	uap = memalloc(sizeof(opt_t));
	memset(uap, 0, sizeof(opt_t));
	uap->thiszone = gmt2local(0);
	uap->flags = (LPORTS | FPNTA | FASYS);
	while ((ch = getopt(argc, argv, "a:bc:df:e:hi:lno:pqr:t:u:vw:xB:C:DE:F:HLNOP:\
		ST:X")) != -1)
		switch(ch) {
		case 'a':
			if (strncasecmp(optarg, "all", 4) == 0) {
				g.gl_offs=1;
				glob("*.dmp*", 0, NULL, &g);
				for(cnt = 0; cnt < g.gl_pathc; cnt++) {
					fname = chomp(g.gl_pathv[cnt]);
					disassemble_fname(fname);
				}
			} else {
				fname = chomp(optarg);
				disassemble_fname(fname);
			}
			exit(EX_OK);
			break; /* NOT REACHED */
		case 'b':
			uap->flags |= FDAMO;
			break;
		case 'c':
			uap->cflag = strtoul(optarg, &rem, 0);
			if (*rem || rem == optarg || uap->cflag > LONG_MAX ||
				!uap->cflag)
				errx(EX_USAGE, "invalid packet count" \
				    " specification.");
			break;
		case 'd':
			uap->flags |= FBPFD;
			break;
		case 'e':
			uap->flags |= TIMER;
			readtimespec(optarg);
			break;

		case 'f':
			if ((cmdfp = fopen(optarg, "r")) == NULL)
				errx(EX_NOINPUT, "unable to open rules file.");
			logic_load(cmdbuf, cmdfp);
			fclose(cmdfp);
			break;
		case 'h':
			usage(NULL);
			exit(EX_USAGE);
			break;	/* NOT REACHED */
		case 'i':
			uap->iflag = optarg;
			break;
		case 'l':
			setvbuf(stdout, NULL, _IOLBF, 0);
			break;
		case 'n':
			uap->flags |= FPDNS;
			break;
		case 'o':
			if ((stdout = fopen(optarg, "a+")) == NULL)
				err(1, "fopen(%s) failed", optarg);
			break;
		case 'p':
			uap->flags &= ~LPORTS;
			break;
		case 'q':
			uap->flags &= ~FASYS;
			break;
		case 'r':
			uap->rflag = optarg;
			break;
		case 't':
			signal(SIGALRM, (void *)transmit_sigint);
			set_timer(optarg);
			break;
		case 'u':
			uap->uflag = strtoul(optarg, &rem, 0);
			if (*rem || rem == optarg || uap->uflag > LONG_MAX ||
			    !uap->uflag)
				errx(EX_USAGE, "invalid UID specification.");
			break;
		case 'v':
			printf("%s\n", _VERSION);
			exit(0);
			break;
		case 'w':
			uap->wflag = optarg;
			break;
		case 'x':
			uap->flags |= FDHEX;
			break;
		case 'B':
			uap->Bflag = optarg;
			ln_pattern(uap->Bflag);
			break;
		case 'C':
			uap->Cflag = parseCflag(optarg);
			if (!uap->Cflag)
				errx(1, "size specification out of bounds");
			break;
		case 'D':
			uap->flags |= FDUMP;
			break;
		case 'E':
			uap->Eflag = optarg;
			uap->flags &= ~FRCFD;
			break;
		case 'F':
			uap->Fflag = optarg;
			break;
		case 'H':
			uap->flags |= FJHDR;
			break;
		case 'L':
			uap->flags |= FLINK;
			break;
		case 'O':
			uap->Oflag = 1;
			break;
		case 'P':
#if !defined __POWERPC__ && !defined linux
			uap->Pflag = optarg;
			scan_descriptor_tables(uap->Pflag);
			cmdbuf = (char *)compile_logic_list();
#else
			errx(EX_USAGE, "-P not yet supported on this system.");
#endif
			break;
		case 'S':
			uap->flags |= FSYSL;
			break;
		case 'T':
			uap->Tflag = atoi(optarg);
			uap->flags |= PTRACK;
			break;
		case 'X':
			uap->flags &= ~FPNTA;
			uap->flags |= FPHEX;
			break;
		default:
			usage(NULL);
			exit(EX_USAGE);
		}
	argv += optind;
	argc -= optind;
	if (uap->Cflag && !uap->wflag)
		errx(1, "-C requires -w");
	if (IS_FDAMO(uap->flags) && ((!uap->oflag) && (!uap->wflag)))
		errx(EX_USAGE, "fatal: -b flag requires [-o|-F] flag");
	if ((uap->pflag && uap->fflag) ||
	    (uap->pflag && uap->Pflag) ||
	    (uap->fflag && uap->Pflag))
		errx(EX_USAGE, "fatal: -p -P and -f flags are not " \
			"compatible");
	if (!uap->oflag)
		uap->fd = STDOUT_FILENO;
	if (!uap->iflag)
       		if (!(uap->iflag = pcap_lookupdev(ebuf)))
			errx(EX_USAGE, "fatal: pcap_lookupdev: %s", ebuf);
	handle_signals();
	init_tabs();
	if (!uap->fflag && !uap->Pflag) {
		cmdbuf = retrive_pcap_expr(argv);
		if (cmdbuf == NULL)
			cmdbuf = " ";
	}
	if (uap->rflag != NULL) {
		pd = pcap_open_offline(uap->rflag, ebuf);
		if (pd == NULL)
			errx(EX_DATAERR, "fatal: pcap_open_offline: failed to "
			    "init pcap: %s", ebuf);
		memset(&localnet, 0, sizeof(localnet));
		memset(&netmask, 0, sizeof(netmask));
	} else {
		pd = pcap_open_live(uap->iflag, 65535, IS_FASYS(uap->flags),
		    READ_TIMEOUT, ebuf);
		if (pd == NULL)
			errx(EX_DATAERR, "fatal: pcap_open_live: %s", ebuf);
	}
#if (!defined(linux))
	if (uap->rflag == NULL)
		if (ioctl(pd->fd, BIOCIMMEDIATE, &fg) < 0)
			errx(EX_OSERR, "fatal: ioctl(BIOCIMMEDIATE) failed.");
#endif
	if (uap->wflag != NULL) {
		pdmper = pcap_dump_open(pd, uap->wflag);
		if (pdmper == NULL)
			errx(EX_NOINPUT, "fatal: pcap_dump_open: %s",
	 		    pcap_geterr(pd));
		if (uap->Cflag > 0) {
			info.WFileName = uap->wflag;
			info.pd = pd;
			info.p = pdmper;
		}
	}
	if (!uap->rflag)
		fprintf(stderr, "ipex: open live: listening on: %s\n",
		    uap->iflag);
	else
		fprintf(stderr, "ipex: reading packets from file: %s\n",
		    uap->rflag);
	if ((sn = pcap_snapshot(pd)) < snaplen) {
		fprintf(stderr,"snaplen raised from %d to %d", snaplen, sn);
		snaplen = sn;
	}
	if (pcap_lookupnet(uap->iflag, &localnet.s_addr, &netmask.s_addr,
		ebuf) < 0) {
		memset(&localnet, 0, sizeof(localnet));
		memset(&netmask, 0, sizeof(netmask));
		fprintf(stderr,"warning: pcap_lookupnet() failed\n");
	}
	if (uap->fflag)
		prescan_pcap_expr(cmdbuf);
	if (pcap_compile(pd, &bpfcode, cmdbuf, uap->Oflag,
		netmask.s_addr) < 0)
		errx(EX_DATAERR, "fatal: pcap_compile: %s", pcap_geterr(pd));
	if (IS_FBPFD(uap->flags)) {
		dump_struct_bpf_program(&bpfcode);
		return(0);
	}
	if (pcap_setfilter(pd, &bpfcode) < 0)
		errx(EX_DATAERR, "fatal: pcap_setfilter() failed");
	if ((uap->loffset = datalink_lookup_offset(pd)) == -1)
		errx(EX_UNAVAILABLE, "fatal: unsupported interface type");
	if (IS_FDAMO(uap->flags)) {
		if (getppid() == 0)
			goto init;
		switch(pid = fork()) {
		case -1:
			errx(EX_OSERR, "unable to fork into background.");
			break;	/* NOT REACHED */
		case 0:
			if ((pidfd = open("/var/run/ipex.pid", O_CREAT
			    | O_WRONLY | O_TRUNC | O_NONBLOCK, 0644)) < 0)
				syslog(LOG_NOTICE, "unable to open pid file.");
			inet_write(pidfd, "%d", getpid());
			close(pidfd); close(0); close(1); close(2);
			if (uap->uflag != '\0') {
				if (setuid(uap->uflag) < 0)
					syslog(LOG_NOTICE, "unable to set "
					    "userid: %d.", uap->uflag);
				else
					fprintf(stderr, "ipex: notice: set "
					    "uid to (%d)\n\r", uap->uflag);
			}
			if (uap->wflag != NULL) {
				pargs = (u_char *)pdmper;
				if (pcap_loop(pd, uap->cflag,
					(pcap_handler)pcap_dump, pargs) < 0) {
					syslog(LOG_NOTICE, "pcap loop failed "
						"unexpectedly?");
					_exit(1);
				}
			} else {
				if (pcap_loop(pd, uap->cflag, (pcap_handler)
					delegate, NULL) < 0) {
					syslog(LOG_NOTICE, "cap loop failed "
						"unexpectedly?");
					_exit(1);
				}
			}
			pcap_stats(pd, &pstat);
			fprintf(stderr,"\nsession stats: %d packets "
			    "received, %d dropped\n", pstat.ps_recv,
			    pstat.ps_drop);
			pcap_close(pd);
			free(uap);
			_exit(0);
			break;	/* NOT REACHED */
		default:
			fprintf(stderr,"ipex: daemonized running with process "
			    "id (%d)\n", pid);
		}
		return(0);
	}
	goto init;
init:
	if (uap->wflag != NULL) {
		pcap_handler printer = pcap_dump;
		pargs = (u_char *)pdmper;
		if (uap->Cflag > 0) {
			printer = dump_and_trunc;
			pargs = (caddr_t)&info;
		}
		if (pcap_loop(pd, uap->cflag, printer, (caddr_t)pargs) < 0) {
			syslog(LOG_NOTICE, "pcap loop failed "
			    "unexpectedly?");
			_exit(1);
		}
	} else {
		if (pcap_loop(pd, uap->cflag, (pcap_handler)
			delegate, NULL) < 0) {
			syslog(LOG_NOTICE, "cap loop failed "
			    "unexpectedly?");
			_exit(1);
		}
	}
	if (uap->rflag == NULL) {
		pcap_stats(pd, &pstat);
		fprintf(stderr, "\nsession statistics: %d packets"
		    "received, %d dropped %d written.\n", pstat.ps_recv,
		    pstat.ps_drop, uap->pcount);
	} else {
		fprintf(stderr, "diagnostic: packet dump completed without "
		    "error: file: %s\n", uap->rflag);
	}
	pcap_close(pd);
	free(uap);
	return(0);
}

void 
cleanup(void)
{
	struct pcap_stat pstat;

	pcap_stats(pd, &pstat);
	fprintf(stderr, "\ninterrupted: session stats: %d packets received, "
	    "%d dropped, %d written.\n",
	    (pstat.ps_recv), (pstat.ps_drop), (uap->pcount));
	close(uap->fd);
	memset(uap, 0, sizeof(opt_t));
	free(uap);
	free(protocols);
	free(tcp_ports);
	free(udp_ports);
	flush_ns_cache();
	if (uap->flags & PTRACK)
		heap_destory(conqueue);
	pcap_close(pd);
	exit(0);
}

int  
get_proto_printer(u_char proto, int caplen)
{
	const struct proto_printers *p;

	for (p = p_funcs; p->pp_printer; p++)
		if (proto == p->pp_proto) {
			/*
			 * extra check here to make sure we have enough
			 * data from the wire to align the proper protocol
			 * header later.
			 */
			if (caplen >= p->pp_hlen)
				return((int)p->pp_printer);
			else
				break;
		}

	return(NO_PRINT);
}

int
datalink_lookup_offset(pcap_t *p)
{
	switch(pcap_datalink(p)) {
	case DLT_EN10MB:
	case DLT_IEEE802:
		return(ETHHDR_SIZE);
	case DLT_FDDI:
		return(FDDIHDR_SIZE);
	case DLT_SLIP:
		return(SLIPHDR_SIZE);
	case DLT_PPP:
		return(PPPHDR_SIZE);
	case DLT_RAW:
		return(RAWHDR_SIZE);
#ifdef HAVE_NET_IF_PFLOG_H
	case DLT_PFLOG:
		return(PFLOG_HDRLEN);
#endif
	case DLT_NULL:
		return(LOOPHDR_SIZE);
	default:
		break;
	}
	return(-1);	/* unknown datalink layer type */
}

void 
delegate(u_char *pargs, register struct pcap_pkthdr *h, u_char *p)
{
	register struct ip *ip = (struct ip *)(p + uap->loffset);
	register int (*print_pointer)
			(u_char *, struct pcap_pkthdr *, u_char *);
	int status = 0;
	char ptr1[65535];
	if (uap->Eflag) {
		int len = h->len - uap->loffset;

		memset(&ptr1[0], 0, len);
		memcpy(ptr1, ip, len);
		rettext(ptr1, len);
		if (grep(uap->Eflag, ptr1, uap->fd) != 0)
			return;

	}
	if (ip->ip_p != IPPROTO_TCP && uap->flags & PTRACK)
		return;
	if (uap->flags & TIMER) {
		if ((((u_int)h->ts.tv_sec) >= ((u_int)dsp->lower)) &&
			((u_int)(h->ts.tv_sec) <= (u_int)(dsp->upper)))
			status = 1;
	}
	if (!status && (uap->flags & TIMER))
		return;
	if (IS_FDUMP(uap->flags)) {
		ascii_print_with_offset(p, h->len, 0x100, uap->fd);
		return;
	}
	if (uap->Bflag != NULL)
		if (!detectpattern(p, h->len))
			return;
	if (uap->flags & PTRACK) {
		handle_sessions(h, p);
		return;
   	}
	if (((int)print_pointer = get_proto_printer(ip->ip_p, h->caplen)) ==
		NO_PRINT) {
		(int)print_pointer = generic_print;
		(*print_pointer)((u_char *)uap, h, p);
		return;
	}
	(*print_pointer)((u_char *)uap, h, p);
	return;
}

int
usage(base)
        char *base;
{
	fprintf(stderr,
	"usage: ipex [-hvnqbHxTLGdO] [-f config] [-r pattern] [-o outfile]\n"
	"            [-c count] [-i interface] [-u uid] [-P op=arg]\n"
	"	    [-t time] [-w file] [-F file] [expression] [-B file]\n");
	return (1);
}

#if (defined(HAVE___ATTRIBUTE__))
int version(void)
    __attribute__ ((__noreturn__));
#endif	/* HAVE___ATTRIBUTE__ */

void
write_current_statistics(void)
{
	int sfd;
	struct pcap_stat pstat;

	pcap_stats(pd, &pstat);
	if ((sfd = open("ipex.stats", O_WRONLY|O_CREAT)) < 0) {
		fprintf(stderr, "WARNING: failed to create ipex stats\n");
		return;
	}
	inet_write(sfd,
	    "Current ipex statistics:\n"
	    "ipex recieved %d packets off interface /dev/%s\n"
	    "out of those, %d were dropped and\n"
	    "%d have been written.\n",
	    (pstat.ps_recv), (uap->iflag),
	    (pstat.ps_drop), (uap->pcount));
	close(sfd);
	return;
}

void
init_tabs(void)
{
	struct protoent *p;
	struct  servent *s;
	char *name, **tab;
	int port;

	if (protocols != NULL) {
		free(protocols);
		protocols = NULL;
	}
	protocols = memalloc(256 * sizeof(*protocols));
	memset(protocols, 0, 256 * sizeof(*protocols));
	setprotoent(1);
	while ((p = getprotoent()) != NULL)
		if (p->p_proto >= 0 && p->p_proto <= 255 && p->p_name != NULL)
			protocols[p->p_proto] = strdup(p->p_name);
	endprotoent();
	if (udp_ports != NULL) {
		free(udp_ports);
		udp_ports = NULL;
	}
	udp_ports = memalloc(65536 * sizeof(*udp_ports));
	if (udp_ports != NULL)
		memset(udp_ports, 0, 65536 * sizeof(*udp_ports));
	if (tcp_ports != NULL) {
		free(tcp_ports);
		tcp_ports = NULL;
	}
	tcp_ports = memalloc(65536 * sizeof(*tcp_ports));
	if (tcp_ports != NULL)
		memset(tcp_ports, 0, 65536 * sizeof(*tcp_ports));
	setservent(1);
	while ((s = getservent()) != NULL) {
		if (s->s_proto == NULL)
			continue;
		else if (!strcmp(s->s_proto, "tcp")) {
			port = ntohs(s->s_port);
			name = s->s_name;
			tab = tcp_ports;
		} else if (!strcmp(s->s_proto, "udp")) {
			port = ntohs(s->s_port);
			name = s->s_name;
			tab = udp_ports;
		} else
			continue;
		if ((port < 0 || port > 65535) || (name == NULL))
			continue;
		tab[port] = strdup(name);
	}
	return;
}

char *
getproto(u_int p)
{
	static char pnum[4];
	char *s;

	p &= 0xff;
	s = protocols ? protocols[p] : NULL;
	if (s == NULL) {
		sprintf(pnum, "%u", p);
		s = pnum;
	}
	return (s);
}

char *
portname(int res, char *proto, u_int port)
{
	static char pname[8];
	char *s;

	port = ntohs(port);
	port &= 0xffff;
	sprintf(pname, "%u", port);
	if (uap->flags & FPDNS)
		return pname;
	s = NULL;
	if (!strcmp(proto, "tcp"))
		s = tcp_ports[port];
	else if (!strcmp(proto, "udp"))
		s = udp_ports[port];
	if (s == NULL)
		s = pname;
	return (s);
}

/* EOF IPEX_C_ */
